@@ -49,6 +49,21 @@ enum mgcp_connection_mode {
 	MGCP_CONN_LOOPBACK  = 4 | MGCP_CONN_RECV_SEND,
 };
 
+#define MGCP_X_OSMO_IGN_HEADER "X-Osmo-IGN:"
+#define MGCP_X_OSMO_OSMUX_HEADER "X-Osmux:"
+
+/* Values should be bitwise-OR-able */
+enum mgcp_x_osmo_ign {
+	MGCP_X_OSMO_IGN_NONE = 0,
+	MGCP_X_OSMO_IGN_CALLID = 1,
+};
+
+/* Codec parameters (communicated via SDP/fmtp) */
+struct mgcp_codec_param {
+	bool amr_octet_aligned_present;
+	bool amr_octet_aligned;
+};
+
 /* Ensure that the msg->l2h is NUL terminated. */
 static inline int mgcp_msg_terminate_nul(struct msgb *msg)
 {
@@ -71,9 +86,12 @@ static inline int mgcp_msg_terminate_nul(struct msgb *msg)
 /* Maximum length of the comment field */
 #define MGCP_COMMENT_MAXLEN 256
 
-/* String length of Connection Identifiers
- * (see also RFC3435 2.1.3.2 Names of Connections) */
-#define MGCP_CONN_ID_LENGTH 32+1
+/* Maximum allowed String length of Connection Identifiers as per spec
+ * (see also RFC3435 2.1.3.2 Names of Connections), plus one for '\0'. */
+#define MGCP_CONN_ID_MAXLEN 32+1
+
+/* Deprecated: old name of MGCP_CONN_ID_MAXLEN. */
+#define MGCP_CONN_ID_LENGTH MGCP_CONN_ID_MAXLEN
 
 /* String length of Endpoint Identifiers.
 /  (see also RFC3435 section 3.2.1.3) */
