@@ -30,7 +30,7 @@
 
 #include <osmocom/mgcp_client/mgcp_client.h>
 
-#define MGW_STR "Configure MGCP connection to Media Gateway\n"
+#define MGW_STR MGCP_CLIENT_MGW_STR
 
 void *global_mgcp_client_ctx = NULL;
 struct mgcp_client_conf *global_mgcp_client_conf = NULL;
@@ -43,8 +43,9 @@ DEFUN(cfg_mgw_local_ip, cfg_mgw_local_ip_cmd,
 	if (!global_mgcp_client_conf)
 		return CMD_ERR_NOTHING_TODO;
 	OSMO_ASSERT(global_mgcp_client_ctx);
-	global_mgcp_client_conf->local_addr =
-		talloc_strdup(global_mgcp_client_ctx, argv[0]);
+	osmo_talloc_replace_string(global_mgcp_client_ctx,
+				   (char**)&global_mgcp_client_conf->local_addr,
+				   argv[0]);
 	return CMD_SUCCESS;
 }
 ALIAS_DEPRECATED(cfg_mgw_local_ip, cfg_mgcpgw_local_ip_cmd,
@@ -75,8 +76,9 @@ DEFUN(cfg_mgw_remote_ip, cfg_mgw_remote_ip_cmd,
 	if (!global_mgcp_client_conf)
 		return CMD_ERR_NOTHING_TODO;
 	OSMO_ASSERT(global_mgcp_client_ctx);
-	global_mgcp_client_conf->remote_addr =
-		talloc_strdup(global_mgcp_client_ctx, argv[0]);
+	osmo_talloc_replace_string(global_mgcp_client_ctx,
+				   (char**)&global_mgcp_client_conf->remote_addr,
+				   argv[0]);
 	return CMD_SUCCESS;
 }
 ALIAS_DEPRECATED(cfg_mgw_remote_ip, cfg_mgcpgw_remote_ip_cmd,
@@ -99,23 +101,14 @@ ALIAS_DEPRECATED(cfg_mgw_remote_port, cfg_mgcpgw_remote_port_cmd,
 		 MGW_STR "remote bind to connect to MGCP gateway with\n"
 		 "remote bind port\n")
 
-DEFUN(cfg_mgw_endpoint_range, cfg_mgw_endpoint_range_cmd,
+DEFUN_DEPRECATED(cfg_mgw_endpoint_range, cfg_mgw_endpoint_range_cmd,
       "mgw endpoint-range <1-65534> <1-65534>",
-      MGW_STR "usable range of endpoint identifiers\n"
-      "set first usable endpoint identifier\n"
-      "set last usable endpoint identifier\n")
+      MGW_STR "DEPRECATED: the endpoint range cannot be defined by the client\n"
+      "-\n" "-\n")
 {
-	uint16_t first_endpoint = atoi(argv[0]);
-	uint16_t last_endpoint = atoi(argv[1]);
-
-	if (last_endpoint < first_endpoint) {
-		vty_out(vty, "last endpoint must be greater than first endpoint!%s",
-			VTY_NEWLINE);
-		return CMD_SUCCESS;
-	}
-
-	global_mgcp_client_conf->first_endpoint = first_endpoint;
-	global_mgcp_client_conf->last_endpoint = last_endpoint;
+	vty_out(vty, "Please do not use legacy config 'mgw endpoint-range'"
+		" (the range can no longer be defined by the MGCP client)%s",
+		VTY_NEWLINE);
 	return CMD_SUCCESS;
 }
 ALIAS_DEPRECATED(cfg_mgw_endpoint_range, cfg_mgcpgw_endpoint_range_cmd,
@@ -126,14 +119,15 @@ ALIAS_DEPRECATED(cfg_mgw_endpoint_range, cfg_mgcpgw_endpoint_range_cmd,
 
 #define BTS_START_STR "First UDP port allocated for the BTS side\n"
 #define UDP_PORT_STR "UDP Port number\n"
-DEFUN(cfg_mgw_rtp_bts_base_port,
+DEFUN_DEPRECATED(cfg_mgw_rtp_bts_base_port,
       cfg_mgw_rtp_bts_base_port_cmd,
       "mgw bts-base <0-65534>",
       MGW_STR
-      BTS_START_STR
-      UDP_PORT_STR)
+      "DEPRECATED: there is no explicit BTS side in current osmo-mgw\n" "-\n")
 {
-	global_mgcp_client_conf->bts_base = atoi(argv[0]);
+	vty_out(vty, "Please do not use legacy config 'mgw bts-base'"
+		" (there is no explicit BTS side in an MGW anymore)%s",
+		VTY_NEWLINE);
 	return CMD_SUCCESS;
 }
 ALIAS_DEPRECATED(cfg_mgw_rtp_bts_base_port,
@@ -143,13 +137,26 @@ ALIAS_DEPRECATED(cfg_mgw_rtp_bts_base_port,
       BTS_START_STR
       UDP_PORT_STR)
 
+DEFUN(cfg_mgw_endpoint_domain_name,
+      cfg_mgw_endpoint_domain_name_cmd,
+      "mgw endpoint-domain NAME",
+      MGW_STR "Set the domain name to send in MGCP messages, e.g. the part 'foo' in 'rtpbridge/*@foo'.\n"
+      "Domain name, should be alphanumeric.\n")
+{
+	if (osmo_strlcpy(global_mgcp_client_conf->endpoint_domain_name, argv[0],
+			 sizeof(global_mgcp_client_conf->endpoint_domain_name))
+	    >= sizeof(global_mgcp_client_conf->endpoint_domain_name)) {
+		vty_out(vty, "%% Error: 'mgw endpoint-domain' name too long, max length is %zu: '%s'%s",
+			sizeof(global_mgcp_client_conf->endpoint_domain_name) - 1, argv[0], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+	return CMD_SUCCESS;
+}
+
 int mgcp_client_config_write(struct vty *vty, const char *indent)
 {
 	const char *addr;
 	int port;
-	uint16_t first_endpoint;
-	uint16_t last_endpoint;
-	uint16_t bts_base;
 
 	addr = global_mgcp_client_conf->local_addr;
 	if (addr)
@@ -169,18 +176,9 @@ int mgcp_client_config_write(struct vty *vty, const char *indent)
 		vty_out(vty, "%smgw remote-port %u%s", indent,
 			(uint16_t)port, VTY_NEWLINE);
 
-	first_endpoint = global_mgcp_client_conf->first_endpoint;
-	last_endpoint = global_mgcp_client_conf->last_endpoint;
-	if (last_endpoint != 0) {
-		vty_out(vty, "%smgw endpoint-range %u %u%s", indent,
-			first_endpoint, last_endpoint, VTY_NEWLINE);
-	}
-
-	bts_base = global_mgcp_client_conf->bts_base;
-	if (bts_base) {
-		vty_out(vty, "%smgw bts-base %u%s", indent,
-			bts_base, VTY_NEWLINE);
-	}
+	if (global_mgcp_client_conf->endpoint_domain_name[0])
+		vty_out(vty, "%smgw endpoint-domain %s%s", indent,
+			global_mgcp_client_conf->endpoint_domain_name, VTY_NEWLINE);
 
 	return CMD_SUCCESS;
 }
@@ -196,6 +194,7 @@ void mgcp_client_vty_init(void *talloc_ctx, int node, struct mgcp_client_conf *c
 	install_element(node, &cfg_mgw_remote_port_cmd);
 	install_element(node, &cfg_mgw_endpoint_range_cmd);
 	install_element(node, &cfg_mgw_rtp_bts_base_port_cmd);
+	install_element(node, &cfg_mgw_endpoint_domain_name_cmd);
 
 	/* deprecated 'mgcpgw' commands */
 	install_element(node, &cfg_mgcpgw_local_ip_cmd);
