@@ -22,6 +22,7 @@
  */
 
 #include <osmocom/core/talloc.h>
+#include <osmocom/vty/misc.h>
 #include <osmocom/mgcp/mgcp.h>
 #include <osmocom/mgcp/mgcp_common.h>
 #include <osmocom/mgcp/mgcp_internal.h>
@@ -31,10 +32,13 @@
 
 #include <string.h>
 #include <inttypes.h>
+#include <limits.h>
 
 #define RTCP_OMIT_STR "Drop RTCP packets in both directions\n"
 #define RTP_PATCH_STR "Modify RTP packet header in both directions\n"
 #define RTP_KEEPALIVE_STR "Send dummy UDP packet to net RTP destination\n"
+#define RTP_TS101318_RFC5993_CONV_STR "Convert GSM-HR from TS101318 to RFC5993 and vice versa\n"
+
 
 static struct mgcp_config *g_cfg = NULL;
 
@@ -94,13 +98,17 @@ static int config_write_mgcp(struct vty *vty)
 	else
 		vty_out(vty, "  no rtcp-omit%s", VTY_NEWLINE);
 	if (g_cfg->trunk.force_constant_ssrc
-	    || g_cfg->trunk.force_aligned_timing) {
+	    || g_cfg->trunk.force_aligned_timing
+	    || g_cfg->trunk.rfc5993_hr_convert) {
 		vty_out(vty, "  %srtp-patch ssrc%s",
 			g_cfg->trunk.force_constant_ssrc ? "" : "no ",
 			VTY_NEWLINE);
 		vty_out(vty, "  %srtp-patch timestamp%s",
 			g_cfg->trunk.force_aligned_timing ? "" : "no ",
 			VTY_NEWLINE);
+		vty_out(vty, "  %srtp-patch rfc5993hr%s",
+			g_cfg->trunk.rfc5993_hr_convert ? "" : "no ",
+			VTY_NEWLINE);
 	} else
 		vty_out(vty, "  no rtp-patch%s", VTY_NEWLINE);
 	if (g_cfg->trunk.audio_payload != -1)
@@ -152,6 +160,10 @@ static int config_write_mgcp(struct vty *vty)
 		vty_out(vty, "  osmux dummy %s%s",
 			g_cfg->osmux_dummy ? "on" : "off", VTY_NEWLINE);
 	}
+
+	if (g_cfg->conn_timeout)
+		vty_out(vty, "  conn-timeout %u%s", g_cfg->conn_timeout, VTY_NEWLINE);
+
 	return CMD_SUCCESS;
 }
 
@@ -160,11 +172,19 @@ static void dump_rtp_end(struct vty *vty, struct mgcp_conn_rtp *conn)
 	struct mgcp_rtp_state *state = &conn->state;
 	struct mgcp_rtp_end *end = &conn->end;
 	struct mgcp_rtp_codec *codec = end->codec;
+	struct rate_ctr *tx_packets, *tx_bytes;
+	struct rate_ctr *rx_packets, *rx_bytes;
 	struct rate_ctr *dropped_packets;
 
+	tx_packets = &conn->rate_ctr_group->ctr[RTP_PACKETS_TX_CTR];
+	tx_bytes = &conn->rate_ctr_group->ctr[RTP_OCTETS_TX_CTR];
+	rx_packets = &conn->rate_ctr_group->ctr[RTP_PACKETS_RX_CTR];
+	rx_bytes = &conn->rate_ctr_group->ctr[RTP_OCTETS_RX_CTR];
 	dropped_packets = &conn->rate_ctr_group->ctr[RTP_DROPPED_PACKETS_CTR];
 
 	vty_out(vty,
+		"   Packets Sent: %" PRIu64 " (%" PRIu64 " bytes total)%s"
+		"   Packets Received: %" PRIu64 " (%" PRIu64 " bytes total)%s"
 		"   Timestamp Errs: %" PRIu64 "->%" PRIu64 "%s"
 		"   Dropped Packets: %" PRIu64 "%s"
 		"   Payload Type: %d Rate: %u Channels: %d %s"
@@ -172,6 +192,8 @@ static void dump_rtp_end(struct vty *vty, struct mgcp_conn_rtp *conn)
 		"   FPP: %d Packet Duration: %u%s"
 		"   FMTP-Extra: %s Audio-Name: %s Sub-Type: %s%s"
 		"   Output-Enabled: %d Force-PTIME: %d%s",
+		tx_packets->current, tx_bytes->current, VTY_NEWLINE,
+		rx_packets->current, rx_bytes->current, VTY_NEWLINE,
 		state->in_stream.err_ts_ctr->current,
 		state->out_stream.err_ts_ctr->current,
 	        VTY_NEWLINE,
@@ -184,13 +206,47 @@ static void dump_rtp_end(struct vty *vty, struct mgcp_conn_rtp *conn)
 		end->force_output_ptime, VTY_NEWLINE);
 }
 
-static void dump_trunk(struct vty *vty, struct mgcp_trunk_config *cfg,
-		       int verbose)
+static void dump_endpoint(struct vty *vty, struct mgcp_endpoint *endp, int epidx,
+			  int trunk_nr, enum mgcp_trunk_type trunk_type, int show_stats)
 {
-	int i;
 	struct mgcp_conn *conn;
 
-	vty_out(vty, "%s trunk nr %d with %d endpoints:%s",
+	vty_out(vty, "%s trunk %d endpoint %s%.2x:%s",
+		trunk_type == MGCP_TRUNK_VIRTUAL ? "Virtual" : "E1", trunk_nr,
+		trunk_type == MGCP_TRUNK_VIRTUAL ? MGCP_ENDPOINT_PREFIX_VIRTUAL_TRUNK : "",
+		epidx, VTY_NEWLINE);
+
+	if (llist_empty(&endp->conns)) {
+		vty_out(vty, "   No active connections%s", VTY_NEWLINE);
+		return;
+	}
+
+	llist_for_each_entry(conn, &endp->conns, entry) {
+		vty_out(vty, "   CONN: %s%s", mgcp_conn_dump(conn), VTY_NEWLINE);
+
+		if (show_stats) {
+			if (endp->cfg->conn_timeout) {
+				struct timeval remaining;
+				osmo_timer_remaining(&conn->watchdog, NULL, &remaining);
+				vty_out(vty, "   Currently remaining timeout (seconds): %d.%06d%s",
+					(int)remaining.tv_sec, (int)remaining.tv_usec, VTY_NEWLINE);
+			}
+
+			/* FIXME: Also add verbosity for other
+			 * connection types (E1) as soon as
+			 * the implementation is available */
+			if (conn->type == MGCP_CONN_TYPE_RTP) {
+				dump_rtp_end(vty, &conn->u.rtp);
+			}
+		}
+	}
+}
+
+static void dump_trunk(struct vty *vty, struct mgcp_trunk_config *cfg, int show_stats)
+{
+	int i;
+
+	vty_out(vty, "%s trunk %d with %d endpoints:%s",
 		cfg->trunk_type == MGCP_TRUNK_VIRTUAL ? "Virtual" : "E1",
 		cfg->trunk_nr, cfg->number_endpoints - 1, VTY_NEWLINE);
 
@@ -201,29 +257,35 @@ static void dump_trunk(struct vty *vty, struct mgcp_trunk_config *cfg,
 
 	for (i = 1; i < cfg->number_endpoints; ++i) {
 		struct mgcp_endpoint *endp = &cfg->endpoints[i];
+		dump_endpoint(vty, endp, i, cfg->trunk_nr, cfg->trunk_type, show_stats);
+		if (i < cfg->number_endpoints - 1)
+			vty_out(vty, "%s", VTY_NEWLINE);
+	}
 
-		vty_out(vty, "Endpoint 0x%.2x:%s", i, VTY_NEWLINE);
-
-		llist_for_each_entry(conn, &endp->conns, entry) {
-			vty_out(vty, "   CONN: %s%s",
-				mgcp_conn_dump(conn), VTY_NEWLINE);
-
-			if (verbose) {
-				/* FIXME: Also add verbosity for other
-				 * connection types (E1) as soon as
-				 * the implementation is available */
-				if (conn->type == MGCP_CONN_TYPE_RTP) {
-					dump_rtp_end(vty, &conn->u.rtp);
-				}
-			}
-		}
+	if (show_stats && cfg->mgcp_crcx_ctr_group) {
+		vty_out(vty, "   %s:%s", cfg->mgcp_crcx_ctr_group->desc->group_description, VTY_NEWLINE);
+		vty_out_rate_ctr_group_fmt(vty, "   %25n: %10c (%S/s %M/m %H/h %D/d) %d", cfg->mgcp_crcx_ctr_group);
+	}
+	if (show_stats && cfg->mgcp_dlcx_ctr_group) {
+		vty_out(vty, "   %s:%s", cfg->mgcp_dlcx_ctr_group->desc->group_description, VTY_NEWLINE);
+		vty_out_rate_ctr_group_fmt(vty, "   %25n: %10c (%S/s %M/m %H/h %D/d) %d", cfg->mgcp_dlcx_ctr_group);
+	}
+	if (show_stats && cfg->mgcp_mdcx_ctr_group) {
+		vty_out(vty, "   %s:%s", cfg->mgcp_mdcx_ctr_group->desc->group_description, VTY_NEWLINE);
+		vty_out_rate_ctr_group_fmt(vty, "   %25n: %10c (%S/s %M/m %H/h %D/d) %d", cfg->mgcp_mdcx_ctr_group);
+	}
+	if (show_stats && cfg->all_rtp_conn_stats) {
+		vty_out(vty, "   %s:%s", cfg->all_rtp_conn_stats->desc->group_description, VTY_NEWLINE);
+		vty_out_rate_ctr_group_fmt(vty, "   %25n: %10c (%S/s %M/m %H/h %D/d) %d", cfg->all_rtp_conn_stats);
 	}
 }
 
+#define SHOW_MGCP_STR "Display information about the MGCP Media Gateway\n"
+
 DEFUN(show_mcgp, show_mgcp_cmd,
       "show mgcp [stats]",
       SHOW_STR
-      "Display information about the MGCP Media Gateway\n"
+      SHOW_MGCP_STR
       "Include Statistics\n")
 {
 	struct mgcp_trunk_config *trunk;
@@ -235,12 +297,78 @@ DEFUN(show_mcgp, show_mgcp_cmd,
 	    dump_trunk(vty, trunk, show_stats);
 
 	if (g_cfg->osmux)
-		vty_out(vty, "Osmux used CID: %d%s", osmux_used_cid(),
+		vty_out(vty, "Osmux used CID: %d%s", osmux_cid_pool_count_used(),
 			VTY_NEWLINE);
 
 	return CMD_SUCCESS;
 }
 
+static void
+dump_mgcp_endpoint(struct vty *vty, struct mgcp_trunk_config *trunk, const char *epname)
+{
+	const size_t virt_prefix_len = sizeof(MGCP_ENDPOINT_PREFIX_VIRTUAL_TRUNK) - 1;
+	unsigned long epidx;
+	char *endp;
+	int i;
+
+	if (strncmp(epname, MGCP_ENDPOINT_PREFIX_VIRTUAL_TRUNK, virt_prefix_len) == 0)
+		epname += virt_prefix_len;
+	errno = 0;
+	epidx = strtoul(epname, &endp, 16);
+	if (epname[0] == '\0' || *endp != '\0') {
+		vty_out(vty, "endpoint name '%s' is not a hex number%s", epname, VTY_NEWLINE);
+		return;
+	}
+	if ((errno == ERANGE && epidx == ULONG_MAX) /* parsed value out of range */
+	    || epidx >= trunk->number_endpoints) {
+		vty_out(vty, "endpoint %.2lx not configured on trunk %d%s", epidx, trunk->trunk_nr, VTY_NEWLINE);
+		return;
+	}
+
+	for (i = 0; i < trunk->number_endpoints; ++i) {
+		struct mgcp_endpoint *endp = &trunk->endpoints[i];
+		if (i == epidx) {
+			dump_endpoint(vty, endp, i, trunk->trunk_nr, trunk->trunk_type, true);
+			break;
+		}
+	}
+}
+
+DEFUN(show_mcgp_endpoint, show_mgcp_endpoint_cmd,
+      "show mgcp endpoint NAME",
+      SHOW_STR
+      SHOW_MGCP_STR
+      "Display information about an endpoint\n" "The name of the endpoint\n")
+{
+	struct mgcp_trunk_config *trunk;
+
+	dump_mgcp_endpoint(vty, &g_cfg->trunk, argv[0]);
+	llist_for_each_entry(trunk, &g_cfg->trunks, entry)
+		dump_mgcp_endpoint(vty, trunk, argv[0]);
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(show_mcgp_trunk_endpoint, show_mgcp_trunk_endpoint_cmd,
+      "show mgcp trunk <0-64> endpoint NAME",
+      SHOW_STR
+      SHOW_MGCP_STR
+      "Display information about a trunk\n" "Trunk number\n"
+      "Display information about an endpoint\n" "The name of the endpoint\n")
+{
+	struct mgcp_trunk_config *trunk;
+	int trunkidx = atoi(argv[0]);
+
+	trunk = find_trunk(g_cfg, trunkidx);
+	if (!trunk) {
+		vty_out(vty, "trunk %d not found%s", trunkidx, VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	dump_mgcp_endpoint(vty, trunk, argv[1]);
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_mgcp, cfg_mgcp_cmd, "mgcp", "Configure the MGCP")
 {
 	vty->node = MGCP_NODE;
@@ -600,11 +728,28 @@ DEFUN(cfg_mgcp_no_patch_rtp_ts,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_mgcp_patch_rtp_rfc5993hr,
+      cfg_mgcp_patch_rtp_rfc5993hr_cmd,
+      "rtp-patch rfc5993hr", RTP_PATCH_STR RTP_TS101318_RFC5993_CONV_STR)
+{
+	g_cfg->trunk.rfc5993_hr_convert = true;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_mgcp_no_patch_rtp_rfc5993hr,
+      cfg_mgcp_no_patch_rtp_rfc5993hr_cmd,
+      "no rtp-patch rfc5993hr", NO_STR RTP_PATCH_STR RTP_TS101318_RFC5993_CONV_STR)
+{
+	g_cfg->trunk.rfc5993_hr_convert = false;
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_mgcp_no_patch_rtp,
       cfg_mgcp_no_patch_rtp_cmd, "no rtp-patch", NO_STR RTP_PATCH_STR)
 {
 	g_cfg->trunk.force_constant_ssrc = 0;
 	g_cfg->trunk.force_aligned_timing = 0;
+	g_cfg->trunk.rfc5993_hr_convert = false;
 	return CMD_SUCCESS;
 }
 
@@ -701,13 +846,17 @@ static int config_write_trunk(struct vty *vty)
 			vty_out(vty, "  rtcp-omit%s", VTY_NEWLINE);
 		else
 			vty_out(vty, "  no rtcp-omit%s", VTY_NEWLINE);
-		if (trunk->force_constant_ssrc || trunk->force_aligned_timing) {
+		if (trunk->force_constant_ssrc || trunk->force_aligned_timing
+		    || g_cfg->trunk.rfc5993_hr_convert) {
 			vty_out(vty, "  %srtp-patch ssrc%s",
 				trunk->force_constant_ssrc ? "" : "no ",
 				VTY_NEWLINE);
 			vty_out(vty, "  %srtp-patch timestamp%s",
 				trunk->force_aligned_timing ? "" : "no ",
 				VTY_NEWLINE);
+			vty_out(vty, "  %srtp-patch rfc5993hr%s",
+				trunk->rfc5993_hr_convert ? "" : "no ",
+				VTY_NEWLINE);
 		} else
 			vty_out(vty, "  no rtp-patch%s", VTY_NEWLINE);
 		if (trunk->audio_fmtp_extra)
@@ -873,12 +1022,31 @@ DEFUN(cfg_trunk_no_patch_rtp_ts,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_trunk_patch_rtp_rfc5993hr,
+      cfg_trunk_patch_rtp_rfc5993hr_cmd,
+      "rtp-patch rfc5993hr", RTP_PATCH_STR RTP_TS101318_RFC5993_CONV_STR)
+{
+	struct mgcp_trunk_config *trunk = vty->index;
+	trunk->rfc5993_hr_convert = true;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_trunk_no_patch_rtp_rfc5993hr,
+      cfg_trunk_no_patch_rtp_rfc5993hr_cmd,
+      "no rtp-patch rfc5993hr", NO_STR RTP_PATCH_STR RTP_TS101318_RFC5993_CONV_STR)
+{
+	struct mgcp_trunk_config *trunk = vty->index;
+	trunk->rfc5993_hr_convert = false;
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_trunk_no_patch_rtp,
       cfg_trunk_no_patch_rtp_cmd, "no rtp-patch", NO_STR RTP_PATCH_STR)
 {
 	struct mgcp_trunk_config *trunk = vty->index;
 	trunk->force_constant_ssrc = 0;
 	trunk->force_aligned_timing = 0;
+	trunk->rfc5993_hr_convert = false;
 	return CMD_SUCCESS;
 }
 
@@ -1137,14 +1305,7 @@ DEFUN(cfg_mgcp_osmux,
 	if (strcmp(argv[0], "off") == 0) {
 		g_cfg->osmux = OSMUX_USAGE_OFF;
 		return CMD_SUCCESS;
-	}
-
-	/* Since OSMUX support is not finished, we do not
-	 * allow to turn it on yet. */
-	vty_out(vty, "OSMUX currently unavailable in this software version.%s", VTY_NEWLINE);
-	return CMD_WARNING;
-#if 0
-	if (strcmp(argv[0], "on") == 0)
+	} else if (strcmp(argv[0], "on") == 0)
 		g_cfg->osmux = OSMUX_USAGE_ON;
 	else if (strcmp(argv[0], "only") == 0)
 		g_cfg->osmux = OSMUX_USAGE_ONLY;
@@ -1155,7 +1316,7 @@ DEFUN(cfg_mgcp_osmux,
 	}
 
 	return CMD_SUCCESS;
-#endif
+
 }
 
 DEFUN(cfg_mgcp_osmux_ip,
@@ -1208,15 +1369,32 @@ DEFUN(cfg_mgcp_osmux_dummy,
 
 DEFUN(cfg_mgcp_domain,
       cfg_mgcp_domain_cmd,
-      "domain NAME", "domain\n" "qualified domain name\n")
+      "domain NAME",
+      "Set the domain part expected in MGCP messages' endpoint names\n"
+      "Qualified domain name expected in MGCP endpoint names, or '*' to accept any domain\n")
 {
 	osmo_strlcpy(g_cfg->domain, argv[0], sizeof(g_cfg->domain));
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_mgcp_conn_timeout,
+      cfg_mgcp_conn_timeout_cmd,
+      "conn-timeout <0-65534>",
+      "Set a time after which inactive connections (CIs) are closed. Set to 0 to disable timeout. This can be used to"
+      " work around interoperability problems causing connections to stay open forever, and slowly exhausting all"
+      " available ports. Enable keep-alive packets in MGW clients when using this option together with LCLS (OsmoBSC,"
+      " OsmoMSC: 'rtp keep-alive')!\n"
+      "Timeout value (sec.)\n")
+{
+	g_cfg->conn_timeout = strtoul(argv[0], NULL, 10);
+	return CMD_SUCCESS;
+}
+
 int mgcp_vty_init(void)
 {
 	install_element_ve(&show_mgcp_cmd);
+	install_element_ve(&show_mgcp_endpoint_cmd);
+	install_element_ve(&show_mgcp_trunk_endpoint_cmd);
 	install_element(ENABLE_NODE, &loop_conn_cmd);
 	install_element(ENABLE_NODE, &tap_rtp_cmd);
 	install_element(ENABLE_NODE, &free_endp_cmd);
@@ -1262,6 +1440,8 @@ int mgcp_vty_init(void)
 	install_element(MGCP_NODE, &cfg_mgcp_patch_rtp_ts_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_no_patch_rtp_ts_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_no_patch_rtp_cmd);
+	install_element(MGCP_NODE, &cfg_mgcp_patch_rtp_rfc5993hr_cmd);
+	install_element(MGCP_NODE, &cfg_mgcp_no_patch_rtp_rfc5993hr_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_sdp_fmtp_extra_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_sdp_payload_send_ptime_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_no_sdp_payload_send_ptime_cmd);
@@ -1276,6 +1456,7 @@ int mgcp_vty_init(void)
 	install_element(MGCP_NODE, &cfg_mgcp_allow_transcoding_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_no_allow_transcoding_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_domain_cmd);
+	install_element(MGCP_NODE, &cfg_mgcp_conn_timeout_cmd);
 
 	install_element(MGCP_NODE, &cfg_mgcp_trunk_cmd);
 	install_node(&trunk_node, config_write_trunk);
@@ -1292,6 +1473,8 @@ int mgcp_vty_init(void)
 	install_element(TRUNK_NODE, &cfg_trunk_patch_rtp_ssrc_cmd);
 	install_element(TRUNK_NODE, &cfg_trunk_no_patch_rtp_ssrc_cmd);
 	install_element(TRUNK_NODE, &cfg_trunk_patch_rtp_ts_cmd);
+	install_element(TRUNK_NODE, &cfg_trunk_patch_rtp_rfc5993hr_cmd);
+	install_element(TRUNK_NODE, &cfg_trunk_no_patch_rtp_rfc5993hr_cmd);
 	install_element(TRUNK_NODE, &cfg_trunk_no_patch_rtp_ts_cmd);
 	install_element(TRUNK_NODE, &cfg_trunk_no_patch_rtp_cmd);
 	install_element(TRUNK_NODE, &cfg_trunk_sdp_fmtp_extra_cmd);
