@@ -104,7 +104,7 @@ static void handle_options(int argc, char **argv)
 			{0, 0, 0, 0},
 		};
 
-		c = getopt_long(argc, argv, "hc:VD", long_options, &option_index);
+		c = getopt_long(argc, argv, "hc:sVD", long_options, &option_index);
 
 		if (c == -1)
 			break;
@@ -132,6 +132,10 @@ static void handle_options(int argc, char **argv)
 			break;
 		};
 	}
+	if (argc > optind) {
+		fprintf(stderr, "Unsupported positional arguments on command line\n");
+		exit(2);
+	}
 }
 
 /* Callback function to be called when the RSIP ("Reset in Progress") mgcp
@@ -270,9 +274,9 @@ int main(int argc, char **argv)
 
 	vty_info.copyright = osmomgw_copyright;
 	vty_init(&vty_info);
-	logging_vty_add_cmds(NULL);
+	logging_vty_add_cmds();
 	osmo_talloc_vty_add_cmds();
-	osmo_stats_vty_add_cmds(&log_info);
+	osmo_stats_vty_add_cmds();
 	mgcp_vty_init();
 
 	handle_options(argc, argv);
